/*************************************************************************
 * TLV 02/14/2002                                                        *
 *                                                                       *
 * This program is for programming the I-PAC keyboard encoder            *
 * For more information, see http://www.ultimarc.com                     *
 *                                                                       *
 * This is free software; you can redistribute it and/or modify          *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation; either version 2 of the License, or     *
 * at your option) any later version.                                    *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with Foobar; if not, write to:                                  *
 *                   Free Software Foundation, Inc.                      *
 *                   59 Temple Place                                     *
 *                   Suite 330                                           *
 *                   Boston, MA  02111-1307                              *
 *                   USA                                                 *
 *************************************************************************/


#include "ipac_prog.h"
#include "kbd_util.h"
#include "term_util.h"
#include "util.h"

/************************************************************
 *Send 1 byte to the I-PAC connected to the keyboard port   *
 *This is done in a sorta funky way...                      *
 *Since we only have 3 bits to work with, we send a byte    *
 *in 3 steps.  First we send the 3 LSB, then middle 3, then *
 *2 MSB + parity.  Currently, parity is ignored by I-PAC    *
 *   Scroll Lock : Bits 0,3,6                               *
 *   Num Lock    : Bits 1,4,7                               *
 *   Caps Lock   : Bits 2,5,P                               *
 *                                                          *
 *   Sequence: Send 0,1,2                                   *
 *             Send 3,4,5                                   *
 *             Send 6,7,P                                   *
 *                                                          *
 * Returns: 1 successfully sent                             *
 *          0 some error sending                            *
 ************************************************************/
int SendByteToIPAC(int console_fd, char byte) {

   int parity;
   int bit;
   int i;
   char cur_byte;

   /* Calculate parity */
   parity = 0;
   for (i=0; i<8; i++) {
      bit=(byte >> i) & 0x01;
      parity = bit ^ parity;
   }
   /*shift parity bit to bit 2 position*/
   parity = parity << 2;

   /* Now, send the byte in 3 passes */
   for (i=0; i<3; i++) {
      switch (i) {
      case 0:
         /* bottom (LSB) 3 bits */
         cur_byte = byte & 0x07;
         break;
      case 1:
         /* Bits 3,4,5 -> bottom 3 bits*/
         cur_byte = (byte & 0x38) >> 3;
         break;
      case 2:
         /* Bits 6,7, Parity -> bottom 3 bits*/
         cur_byte = ((byte & 0xC0) >> 6) | parity;
         break;
      }

      /* Now, actually send the 3 bit chunk */
      if( SendLed(console_fd, cur_byte) != 0) {
         return 0;
      }
   }

   /*Normal completion*/
   usleep(INTER_SEND_DELAY);
   return 1;

}


/************************************************
 *Try to set the I-PAC to programming mode      *
 *This is done in 2 steps and 6 LED set calls   *
 * Step 1) Send anything BUT all LEDs on        *
 * Step 2) Send all LEDs on 5 times in a row    *
 * I-PAC responds with a key press:             *
 *      Y - Ok... in programming mode           *
 *      J - jumper set to mame, can't program   *
 * If the I-PAC doesn't respond in 500ms, error *
 *                                              *
 * Return Value:                                *
 * PGM_MODE_OK - I-PAC in programing mode       *
 * PGM_MODE_FAIL_JUMPER - I-PAC failed to enter *
 *                   programming mode because   *
 *                   jumper is set to MAME      *
 * PGM_MODE_FAIL_TIMEOUT - I-PAC failed to enter*
 *                    programming mode - timeout*
 *                    waiting for response      *
 * PGM_MODE_FAIL_MISC - I-PAC failed to enter   *
 *                 programming mode for some    *
 *                 unknown reason               *  
 *                                              *
 * TLV 02/16/2002 - Add mode parameter for      *
 * programming a second board                   *
 * The sequence for programming board 2 is      *
 * the same as above, except step 2 is:         *
 * Step  2) Send (Scroll & NUM Lock) 5 times    *
 ************************************************/
int SetIPACToProgramMode(int console_fd, int mode, int verbosity) {

   char state_code=LED_ALL_OFF;
   int stdin_fd;
   int i;
   char buf[1];
   int success=0;
   struct termios tty, otty;

   /*Prepare for receiving a character back from the I-PAC
     by opening stdin for reading */
   /*First, turn off line buffering and echoing*/
   /*Save the current terminal settings*/
   if (tcgetattr(STDIN_FILENO, &otty) < 0) {
      perror( "Failed to get current terminal settings");
      return PGM_MODE_FAIL_MISC;
   }

   /*Turn off buffering/echoing*/
   tty = otty;
   if (cbreak(tty) < 0) {
      perror("Failed to disable buffering/echoing\n");
      /*Just to be sure, restore old settings*/
      tcsetattr(STDIN_FILENO, TCSANOW, &otty);
      return PGM_MODE_FAIL_MISC;
   }
   
   /*Open stdin for reading... non-blocking*/
   if ((stdin_fd = open(STDIN_DEV, O_RDONLY | O_NONBLOCK)) == 0) {
      perror( "Failed to open stdin\n");
      return PGM_MODE_FAIL_MISC;
   }

   /*Ok, now stdin is open for reading... try to send the
     enter programming mode code to I-PAC*/


   /*Send 1 all off, then 5 all on*/
   for(i=0; i<6; i++) {
      usleep(INTER_SEND_DELAY);
      SendLed(console_fd, state_code);
      switch (mode) {
      case PROGRAM_BOARD_1:
         /*to program board 1, we send all led on 5 times*/
         state_code = LED_ALL_ON;
         break;
      case PROGRAM_BOARD_2:
         /*to program board 2, we send scr,num on 5 times*/
         state_code = (LED_SCRLCK | LED_NUMLCK);
         break;
      default:
         /*uhhhh bad mode...*/
         printe("Bad programming mode selected: %d\n", mode);
         return PGM_MODE_FAIL_MISC;
      }
       
   }

   /*Ok, programming code sent to I-PAC*/

   /*Now, wait 500ms for return code*/
   /*Try to read the character, then wait 50ms - 10 time*/
   /*Initialize the input buffer*/
   buf[0] = '\0';

   for(i=0; i<10; i++) {
      if (read(stdin_fd, buf, 1) > 0) {
         printv(verbosity, VERBOSE_NORMAL, 
	        "Received response from I-PAC: %c (val: %d)\n", 
	        buf[0], buf[0]);

         /* TLV 02/15/2002 - dunno why, but looks like it's sending me *
          * a 0x10 (linefeed) and _then_ a 0x79 ('y')                  */
         if (buf[0] == 10) {
            printv(verbosity, VERBOSE_VERBOSE, "Caught a chr(10).  Ignoring...\n");
         } else {
            break;
         }
      }
      usleep(MS_DELAY_50);
   }

   /*Close stdin and switch back to normal terminal mode*/
   close(stdin_fd);
   tcsetattr(STDIN_FILENO, TCSANOW, &otty);
  
   /* See what the response from the I-PAC was*/
   switch (buf[0]) {
   case IPAC_RSP_PGM_OK:
      /* I-PAC responded with success */
      return PGM_MODE_OK;
      break;

   case IPAC_RSP_JUMPER_ERR:
      /* I-PAC responded with jumper set to MAME */
      return PGM_MODE_FAIL_JUMPER;
      break;

   case '\0':
      /* buffer unchanged - must have timed out waiting for response */
      return PGM_MODE_FAIL_TIMEOUT;
      break;

   default:
      /* We _shouldn't_ ever hit this... */
      return PGM_MODE_FAIL_MISC;
   }
}

/****************************************************************
 * Try to detect IPAC boards connected to the keyboard port     *
 * To do this, we'll try to put both boards (1&2) in programming*
 * mode.  If they respond, then send bogus data to make sure    *
 * they are not reprogrammed.                                   *
 *                                                              *
 * Return Value:                                                *
 *    DETECT_NONE - didn't detect any boards                    * 
 *    DETECT_BOARD_1 - detected board 1                         *
 *    DETECT_BOARD_2 - detected board 2                         *
 *    DETECT_BOARD_2 | DETECT_BOARD2 - detected both boards     *
 ****************************************************************/
int IPACDetect(int console_fd, int verbosity) {
   int found = DETECT_NONE;
   int shift_code = 0xE1; /*bad shift code to make programming fail*/
   int bogus_codes[IPAC_NUM_KEYCODES];
   int mode[2] = { PROGRAM_BOARD_1, PROGRAM_BOARD_2 };
   int rc;
   int i;


   /*Generate all 0's for bogus data*/
   for (i=0; i<IPAC_NUM_KEYCODES; i++)
      bogus_codes[i] = 0x00;

   for (i=0; i<2; i++) {

      printv(verbosity, VERBOSE_NORMAL, "Probing for I-PAC %d\n", i+1);

      /*Try to set to programming mode and see if there is any response*/
      if (SetIPACToProgramMode(console_fd, mode[i], verbosity)  == PGM_MODE_OK) {

         /*The board responded -- now send bogus data*/
	 printv(verbosity, VERBOSE_VERBOSE,
	        "Board %d responded.  Sending bogus data\n", i+1);

         if (i == 0) {
            found = found | DETECT_BOARD_1;
	 } else {
            found = found | DETECT_BOARD_2;
	 }

	 /*Now, send bogus data - we expect this to fail with a checksum error*/
         rc = ProgramIPAC(console_fd, bogus_codes, shift_code, verbosity, FORCE_PGM_FAIL);
	 if (rc != PGM_OK) {
	    /*Good... failed as expected*/
	    printv(verbosity, VERBOSE_VERBOSE,
	           "Good.  Board rejected bogus data in detection routine\n");
	 } else {
	    /*Uh-oh... the board said it was reprogrammed... to garbage*/
	    printe("Detection error.  Board %d was re-programmed with garbage\n", i+1);
	    printf("To correct, re-run ipacutil using the default config file\n");
	 }
      } else {
      
         printv(verbosity, VERBOSE_NORMAL, "No response from board %d\n", i+1); 
      }

   }

   return found;
}

/****************************************************************
 * You must put the I-PAC in programming mode before using this *
 * 1) Put in programming mode                                   *
 * 2) Send Shift Code (1 byte - FF for none)                    *
 * 3) Send Key Codes  (64 bytes - 32 unshifted/32 shifted keys) *
 * 4) Send Checksum   (1 byte)                                  *
 * Steps 2-4 send a total of 66 bytes                           *
 *                                                              *
 * The checksum is calculated by adding all 65 bytes together   *
 * in an 8 bit buffer and ignoring carry                        *
 *                                                              *
 *                                                              *
 * Return Value:                                                *
 *    PGM_OK - programming of key codes was successful          *
 *    PGM_FAIL_TIMEOUT - timed out waiting for response         *
 *    PGM_FAIL_MISC - failed for some unknown reason            *
 *    PGM_FAIL_CHECKSUM - failed due to checksum error          *
 *                                                              *
 * Add option to generate bad checksum to make sure the         *
 * programming step fails.  This is to support detection        *
 ****************************************************************/
int ProgramIPAC(int console_fd, int *data_array, int shift_code, int verbosity, 
                int force_bad) {

   char buf[1];
   int stdin_fd;
   struct termios tty, otty;
   int i;
   int checksum=shift_code;

   /*Prepare for receiving a character back from the I-PAC
     by opening stdin for reading */
   /*First, turn off line buffering and echoing*/
   /*Save the current terminal settings*/
   if (tcgetattr(STDIN_FILENO, &otty) < 0) {
      perror("Failed to save terminal settings");
      return PGM_MODE_FAIL_MISC;
   }

   /*Turn off buffering/echoing*/
   tty = otty;
   if (cbreak(tty) < 0) {
      perror( "Failed to disable buffering/echoing");
      /*Just to be sure, restore the 'old' terminal settings*/
      tcsetattr(STDIN_FILENO, TCSANOW, &otty);
      return PGM_MODE_FAIL_MISC;
   }
   
   /*Open stdin for reading... non-blocking*/
   if ((stdin_fd = open(STDIN_DEV, O_RDONLY | O_NONBLOCK)) == 0) {
      perror( "Failed to open stdin");
      return PGM_FAIL_MISC;
   }

   /*Send the shift code*/
   if (SendByteToIPAC(console_fd, shift_code) != 1) {
      printe("Error sending shift code to IPAC: 0x%02x\n", shift_code);
      return PGM_FAIL_MISC;
   }

   /* Now, we need to send the 64 bytes that make up the 64 *
    * programmed key codes (32 unshifted/32 shifted)        */
   for(i=0; i<IPAC_NUM_KEYCODES; i++) {
      if (SendByteToIPAC(console_fd, data_array[i]) != 1) {
         printe("Error sending byte to IPAC. offset: %d, code: 0x%02X\n",
	        i, data_array[i]);
	 return PGM_FAIL_MISC;
      }
      checksum = (checksum + data_array[i]) & 0xFF;
   }

   /*Finally, send the checksum*/

   /*TLV 02/18/2002 - if the user wants us to fail this programming step, 
     then make sure the checksum is bad.  This is useful for I-PAC detection
     without blitzing the good data on the board*/
   if (force_bad == FORCE_PGM_FAIL) 
      checksum = checksum + 1;

   if (SendByteToIPAC(console_fd, checksum) != 1) {
      printe("Error sending checksum to IPAC. Checksum: %d\n", checksum);
      return PGM_FAIL_MISC;
   }

   /* DONE sending data... now wait for response */

   /*Now, wait 500ms for return code*/
   /*Try to read the character, then wait 50ms - 10 time*/
   /*Initialize the input buffer*/
   buf[0] = '\0';

   for(i=0; i<10; i++) {
      if (read(stdin_fd, buf, 1) > 0) {
         printv(verbosity, VERBOSE_VERBOSE, 
	        "Received response from I-PAC: %c (val: %d)\n", 
	        buf[0], buf[0]);
         /* TLV 02/15/2002 - dunno why, but looks like it's sending me *
          * a 0x10 (linefeed) and then a 0x79 ('y')                    */
         if (buf[0] == 10) {
            printv(verbosity, VERBOSE_VERBOSE, "Caught a chr(10).  Ignoring...\n");
         } else {
            break;
         }
      }
      usleep(MS_DELAY_50);
   }

   /* Clean up the terminal */
   close(stdin_fd);
   if (tcsetattr(STDIN_FILENO, TCSANOW, &otty) < 0) {
      perror("Failed to restore terminal settings");
      /*should we return here???*/
   }

   switch (buf[0]) {
   case IPAC_RSP_PGM_OK:
      return PGM_OK;
      /*Don't really need the break*/

   case IPAC_RSP_CHECKSUM_ERR:
      return PGM_FAIL_CHECKSUM;
      /*Don't really need the break*/

   case '\0':
      /* Buffer unchanged - must have timed out */
      return PGM_FAIL_TIMEOUT;
      /*Don't really need the break*/

   default:
      return PGM_FAIL_MISC;
   }

}
