/*************************************************************************
 * TLV 02/14/2002                                                        *
 * Copyright (C) 2002 Travis Veldkamp                                    *
 *                                                                       *
 * This program is for programming the I-PAC keyboard encoder            *
 * For more information, see http://www.ultimarc.com                     *
 *                                                                       *
 * This is free software; you can redistribute it and/or modify          *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation; either version 2 of the License, or     *
 * at your option) any later version.                                    *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with Foobar; if not, write to:                                  *
 *                   Free Software Foundation, Inc.                      *
 *                   59 Temple Place                                     *
 *                   Suite 330                                           *
 *                   Boston, MA  02111-1307                              *
 *                   USA                                                 *
 *************************************************************************/


#include "ipac_prog.h"
#include "term_util.h"
#include "kbd_util.h"
#include "cfg_util.h"
#include "cmd_line.h"
#include "util.h"


static const int MAJOR_VERSION = 0;
static const int MINOR_VERSION = 24;
static const int VERSION_DATE  = 20020314;

/*************************************************************************
 * GLOBAL VARIABLES                                                      *
 *************************************************************************/

/*This is set by the signal handler (sig_handle()) when sig_int is caught*
 *This allows us to clean up before falling out with the terminal        *
 *potentially in some wacky state                                        */
int want_out=0;

/*************************************************************************
 * END GLOBAL VARIABLES                                                  *
 *************************************************************************/



/***********************************************************
 * Simple signal handler to set global variable indicating *
 * that the user wants out, but allows us time to clean up *
 ***********************************************************/
void sig_handle(int signal) {

   if (signal == SIGINT) {
      printf("Exiting.... please wait while cleaning up\n");
      want_out = 1;
   }
}

void PrintWelcome() {
   printf("ipacutil v%d.%02d - %d\n", MAJOR_VERSION, 
           MINOR_VERSION, VERSION_DATE);
   printf("I-PAC keyboard encoder programming utility for Linux\n");
   printf("Copyright (C) 2002 Travis Veldkamp\n\n");
}

void ShowUsage() {
   PrintWelcome();

   printf("Usage: ipacutil [-h] [--help] [-v] [-d] [-n] [-b board_num] [-c config_file] [-w config_file]\n");
   printf("-h | --help          -- print this information\n");
   printf("-v | --verbose       -- turn on verbose mode\n");
   printf("-d | --detect        -- detect connected I-PACs\n");
   printf("-n | --no_program    -- don't program (useful with --detect)\n");
   printf("-b board_num         -- program board specified (1 or 2 - default is 1)\n");
   printf("--board board_num    -- program board specified (1 or 2 - default is 1)\n");
   printf("-c config_file       -- use config file to program the ipac\n");
   printf("--config cfg_file    -- use config file to program the ipac\n");
   printf("-w    config_file    -- generate a default config file called config_file\n");
   printf("--write_cfg cfg_file -- generate a default config file called config_file\n");
   printf("-q | --quiet         -- don't print anything to screen except errors\n");
   printf("The default settings correspond to the function when the MAME jumper is on\n");
   printf("\nThis is free software; see the source for copying conditions.  There is NO\n");
   printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
   
}

int main(int argc, char **argv) {

   char ledval;
   int fd = -1;
   int error=0;
   struct termios tty, otty;
   struct cfg_struct cfg_data;
   int i;
   int cur_pos=0;
   int direction=1; /*1 for left->right, -1 for right->left*/
   int shift_code;
   /*The 64 key codes to program the I-PAC with*/
   int keycodes[IPAC_NUM_KEYCODES]; 
   int rc;
   char *cfg_file;
   struct cmdline_struct settings;

   if (parse_cmdline(argc, argv, &settings) < 0) {
      printf("\n");
      ShowUsage();
      exit(1);
   }

   if (settings.showhelp) {
      ShowUsage();
      exit(1);
   }


   if (getuid()) {
      printe("This program can only be run as root\n");
      exit(1);
   }


   /*Register the signal handler for sig int (ctrl-c)*/
   if (signal(SIGINT, &sig_handle) == SIG_ERR) {
      printv(settings.verbose, VERBOSE_VERBOSE,
             "Error registering new signal handler\n");
      error = 1;
      want_out = 1;
   }

   /* open the console we'll use for reading/setting LEDs*/
   if ((!error) && (!error)) {
      if ((fd = open(console, O_RDONLY)) < 0) {
         printe("Error:  Failed to open console\n");
         error = 1;
         want_out = 1;
      }
   }

   if (settings.run_detection) {
      printv(settings.verbose, VERBOSE_NORMAL, "Attempting to detect I-PACs\n");
      printv(settings.verbose, VERBOSE_NORMAL, "\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "********************************************\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "* Do NOT touch the keyboard until          *\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "* detection has completed                  *\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "********************************************\n");

      rc = IPACDetect(fd, settings.verbose);

      if (rc == DETECT_NONE) {

         printf("Didn't detect any I-PACs\n");
         exit(1);

      } else {

         if (rc & DETECT_BOARD_1) {
	    printf("Detected I-PAC set as board 1\n");
         }
         if (rc & DETECT_BOARD_2) {
	    printf("Detected I-PAC set as board 2\n");
	 }
      }
      printv(settings.verbose, VERBOSE_NORMAL, "Detection complete\n");
   }

   if (settings.no_program) {
      /* User has asked for no programming step*/
      /* This is most likely used when the user wants to detect their board*/
      printv(settings.verbose, VERBOSE_NORMAL, 
             "No programming cycle requested\n");
      printv(settings.verbose, VERBOSE_VERBOSE, "Exiting.\n");
      exit(0);
   }
   
   if (strlen(settings.config_file) == 0) {
      printe("Config file not specified\n\n");
      ShowUsage();
      exit(1);
   } else {
      /*Config file specified... see if they want to write default or
        read specified*/
      if (settings.generate_config_file) {
         if (WriteDefaultCfg(settings.config_file) >= 0) {
            printf("Created new config file: %s\n", settings.config_file);
         } else {
            printe("Failed to create config file: %s\n", settings.config_file);
         }
         exit(1);
      } 
   }
  
   printv(settings.verbose, VERBOSE_NORMAL,
          "Programming I-PAC using %s for configuration\n", 
          settings.config_file);

   if ((!error) && (!want_out)) {
      if (settings.verbose) 
         PrintWelcome();

      printv(settings.verbose, VERBOSE_NORMAL, "\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "********************************************\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "* Do NOT touch the keyboard until          *\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "* programming has completed                *\n");
      printv(settings.verbose, VERBOSE_NORMAL,
             "********************************************\n");

      printv(settings.verbose, VERBOSE_VERBOSE,
             "Using %s for keycode data\n", settings.config_file);

      cfg_data.key_code = keycodes;
      printv(settings.verbose, VERBOSE_VERBOSE,
             "Reading in keycode data\n");

      /* Read in the keycode data */
      if (ReadCfg(settings.config_file, &cfg_data) < 0) {
         printe("Error reading in configuration data\n");
	 error = 1;
	 want_out = 1;
      } else {
         /* Read in the shift code data */
         shift_code = cfg_data.shift_code;
         printv(settings.verbose, VERBOSE_VERBOSE, "Shift code is 0x%02X\n", shift_code);
      }
   }

   if ((!want_out) && (!error)) {
      printv(settings.verbose, VERBOSE_VERBOSE, "Setting I-PAC to programming mode... ");
      
      /*set the IPAC to programming mode*/
      if (settings.program_board == 2) {
         /*program board 2*/
         printv(settings.verbose, VERBOSE_VERBOSE, "Programming board 2\n");
         rc = SetIPACToProgramMode(fd, PROGRAM_BOARD_2, settings.verbose);
      } else {
         /*program board 1*/
         printv(settings.verbose, VERBOSE_VERBOSE, "Programming board 1\n");
         rc = SetIPACToProgramMode(fd, PROGRAM_BOARD_1, settings.verbose);
      }
      switch (rc) {
      case PGM_MODE_OK:
         printv(settings.verbose, VERBOSE_VERBOSE, 
	        "Good.  I-PAC now in programming mode\n");
         break;
      case PGM_MODE_FAIL_JUMPER:
         printe("can't program with jumper set to MAME\n");
	 printf("Move the jumper on the I-PAC board and try again\n");
         error = 1;
         break;
      case PGM_MODE_FAIL_MISC:
         printe("some error setting program mode\n");
	 printf("Something is wacked... check connections and try again\n");
         error = 1;
         break;
      case PGM_MODE_FAIL_TIMEOUT:
         printe("timed out setting program mode\n");
	 printf("Verify that the I-PAC is connected and try again\n");
         error = 1;
         break;
      default:
         printe("setting program mode generated unknown response\n");
         error = 1;
      }
   }

   /* Now, try to program the I-PAC */
   if ((!error) && (!want_out)) {
      printv(settings.verbose, VERBOSE_NORMAL, "Programming I-PAC... " );
  
      rc = ProgramIPAC(fd, keycodes, shift_code, settings.verbose, NORMAL_PROGRAM);
      switch (rc) {
      case PGM_OK:
         printv(settings.verbose, VERBOSE_NORMAL, "Done!  I-PAC has been re-programmed\n");
	 break;

      case PGM_FAIL_CHECKSUM:
         printe("the checksum failed while programming\n");
	 printf("Check connections and try again\n");
	 error = 1;
	 break;

      case PGM_FAIL_MISC:
         printe("the programming failed for some reason\n");
	 printf("Something is strange.  Check connections and try again\n");
	 error = 1;
	 break;

      case PGM_FAIL_TIMEOUT:
         printe("timed out waiting for programming response\n");
	 printf("Verify that the I-PAC is connected and try again\n");
	 error = 1;
	 break;

      default:
         printe("programming I-PAC generated unknown response\n");
	 printf("Something is strange.  Check connections and try again\n");
	 error = 1;
      }
   }

   printv(settings.verbose, VERBOSE_VERBOSE, "Cleaning up...\n");

   /*Clean up*/
   if (fd >= 0) {
      SendLed(fd, LED_ALL_OFF);
      close(fd);
   }

   if (error) {
      printf("** Programming FAILED **\n");
      return 1;
   } else {
      printf("** I-PAC has been reprogrammed **\n");
      return 0;
   }

}
